///////////////////////////////////////////////////////////////////////
// FLIC Plugin for VirtualDub 1.8.5+
// Traduction franaise de Sparad0x
// Copyright (C) 2008 Sparad0x
// Copyright (C) 2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <stdio.h>

#include "vdinputdriver.h"
#include "vdunknown.h"

#include "resource.h"

#define USE_DEFAULT_IP_MODEL

static HMODULE g_hModule;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved) {
	if (fdwReason == DLL_PROCESS_ATTACH) {
		g_hModule = (HMODULE)hinstDLL;
	}
	return TRUE;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileFLIC
//
///////////////////////////////////////////////////////////////////////

typedef struct {
	long			mFilesize;
	unsigned short	mVersion;
	short			mFrames;
	short			mWidth;
	short			mHeight;
	short			mBitsPerPixel;
	unsigned short	mFlags;
	long			mFrameRate;
} FlicHeader;

typedef struct {
	unsigned long	mFrameOffset;
	long			mSize;
} FlicIndexEntry;

typedef struct {
	BITMAPINFOHEADER	bmih;
	RGBQUAD				bmic[256];
} FLICINFOHEADER;


class VideoDecoderFLIC;
#ifndef USE_DEFAULT_IP_MODEL
	class VideoDecoderModelFLIC;
#endif
class VideoSourceFLIC;

class InputFileFLIC : public vdxunknown<IVDXInputFile> {

friend VideoDecoderFLIC;
#ifndef USE_DEFAULT_IP_MODEL
	friend VideoDecoderModelFLIC;
#endif
friend VideoSourceFLIC;

private:
	FILE *mFile;

	FlicHeader mHeader;
	FlicIndexEntry *mIndex;

	static INT_PTR CALLBACK InfoDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

protected:
	const VDXInputDriverContext& mContext;

public:
	InputFileFLIC(const VDXInputDriverContext& context);
	~InputFileFLIC();

	void VDXAPIENTRY Init(const wchar_t *szFile, IVDXInputOptions *opts);
	bool VDXAPIENTRY Append(const wchar_t *szFile);

	bool VDXAPIENTRY PromptForOptions(VDXHWND, IVDXInputOptions **);
	bool VDXAPIENTRY CreateOptions(const void *buf, uint32 len, IVDXInputOptions **);
	void VDXAPIENTRY DisplayInfo(VDXHWND hwndParent);

	bool VDXAPIENTRY GetVideoSource(int index, IVDXVideoSource **);
	bool VDXAPIENTRY GetAudioSource(int index, IVDXAudioSource **);
};


///////////////////////////////////////////////////////////////////////
//
//						VideoDecoderFLIC
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderFLIC : public vdxunknown<IVDXVideoDecoder> {

public:
	VideoDecoderFLIC(InputFileFLIC *const pp);
	~VideoDecoderFLIC();

	const void *VDXAPIENTRY DecodeFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, sint64 streamFrame, sint64 targetFrame);
	uint32		VDXAPIENTRY GetDecodePadding();
	void		VDXAPIENTRY Reset();
	bool		VDXAPIENTRY IsFrameBufferValid();
	const VDXPixmap& VDXAPIENTRY GetFrameBuffer();
	bool		VDXAPIENTRY SetTargetFormat(int format, bool useDIBAlignment);
	bool		VDXAPIENTRY SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih);

	const void *VDXAPIENTRY GetFrameBufferBase();
	bool		VDXAPIENTRY IsDecodable(sint64 sample_num);

protected:
	VDXPixmap	mPixmap;
	unsigned char *mFrameBuffer;

private:
	InputFileFLIC *const parentPtr;

	typedef struct {
		BYTE	yuvY;
		BYTE	yuvCb;
		BYTE	yuvCr;
		BYTE	yuvReserved;
	} YUVQUAD;

	RGBQUAD mPalette[256];
	YUVQUAD yPalette[256];

	unsigned char *m8BitBuffer;

	int mBytesPerLine;
	int mPrevFrame;

	const unsigned char *flicbytes;
	unsigned long curbyte;
	unsigned long totbytes;

	int flicGetByte();
	int flicGetWord();
	long flicGetDword();

	void flicYCbCr();
	void flicPalette(bool adjust);
	void flicDelta();
	void flicLC();
	void flicBlack();
	void flicBrun();
	void flicRaw();

	void flicConvertRGB15();
	void flicConvertRGB16();
	void flicConvertRGB24();
	void flicConvertRGB32();

	void flicConvertY8();
	void flicConvertUYVY();
	void flicConvertYUY2();
	void flicConvertYV12();
};

VideoDecoderFLIC::VideoDecoderFLIC(InputFileFLIC *const pp)
	: parentPtr(pp)
{
	int i;
	parentPtr->AddRef();

	mBytesPerLine = (parentPtr->mHeader.mWidth + 3) & 0xFFFFFFFC;
	mFrameBuffer = (unsigned char *)malloc(parentPtr->mHeader.mWidth * parentPtr->mHeader.mHeight * 4);
	m8BitBuffer = (unsigned char *)malloc(mBytesPerLine * parentPtr->mHeader.mHeight);

	// Grayscale palette
	for (i = 0; i < 256; ++i) {
		mPalette[i].rgbBlue = i;
		mPalette[i].rgbGreen = i;
		mPalette[i].rgbRed = i;
		mPalette[i].rgbReserved = 255;
	}
	flicYCbCr();

	mPrevFrame = -1;
	flicbytes = NULL;
	totbytes = 0;
	curbyte = 0;
}

VideoDecoderFLIC::~VideoDecoderFLIC() {
	free(m8BitBuffer);
	free(mFrameBuffer);
	parentPtr->Release();
}

int VideoDecoderFLIC::flicGetByte() {
	int ret = 0;
	if (flicbytes != NULL) {
		if (curbyte < totbytes) {
			ret = flicbytes[curbyte++];
		}
	}
	return ret;
}

int VideoDecoderFLIC::flicGetWord() {
	int ret = flicGetByte();
	return (flicGetByte() << 8) + ret;
}

long VideoDecoderFLIC::flicGetDword() {
	long ret = flicGetWord();
	return (flicGetWord() << 16) + ret;
}

void VideoDecoderFLIC::flicYCbCr() {
	// Convert RGB palette to YUV
	int i;

	for (i = 0; i < 256; ++i) {
		const double B = mPalette[i].rgbBlue;
		const double G = mPalette[i].rgbGreen;
		const double R = mPalette[i].rgbRed;

		yPalette[i].yuvY  = (BYTE)( (0.257 * R) + (0.504 * G) + (0.098 * B) +  16.5);
		yPalette[i].yuvCb = (BYTE)(-(0.148 * R) - (0.291 * G) + (0.439 * B) + 128.5);
		yPalette[i].yuvCr = (BYTE)( (0.439 * R) - (0.368 * G) - (0.071 * B) + 128.5);
	}
}

void VideoDecoderFLIC::flicPalette(bool adjust) {
	int NumPacks, NumColors, index;

	index = 0;
	NumPacks = flicGetWord();

	while (NumPacks > 0) {
		index += flicGetByte();
		if (index > 255) break;

		NumColors = flicGetByte();
		if (NumColors == 0) NumColors = 256;

		do {
			mPalette[index].rgbRed = (BYTE)flicGetByte();
			mPalette[index].rgbGreen = (BYTE)flicGetByte();
			mPalette[index].rgbBlue = (BYTE)flicGetByte();

			if (adjust) {
				// Convert [0...63] to [0...255]

				mPalette[index].rgbBlue = (BYTE)
					((mPalette[index].rgbBlue * 255) / 63);

				mPalette[index].rgbGreen = (BYTE)
					((mPalette[index].rgbGreen * 255) / 63);

				mPalette[index].rgbRed = (BYTE)
					((mPalette[index].rgbRed * 255) / 63);
			}

			++index;
			if (index > 255) {
				// Forced exit
				NumPacks = 0;
				break;
			}

			--NumColors;
		} while (NumColors);
	}

	flicYCbCr();
}

void VideoDecoderFLIC::flicDelta() {
	int i, x, y, LastPixel;
	int NumLines, NumPacks;
	unsigned char *dst;

	NumLines = flicGetWord();
	for (y = 0; NumLines > 0; --NumLines, ++y) {

		LastPixel = -1;
		for (;;) {
			NumPacks = flicGetWord();

			// Opcode coded by the top two bits
			//	00xxxxxx xxxxxxxx     x = NumPacks
			//	01...... ........     (undefined)
			//	10...... xxxxxxxx     x = last pixel of line
			//	11xxxxxx xxxxxxxx     ABS(opcode) is skip value

			i = NumPacks & 0xC000;
			if (i == 0) break;
			if (i == 0x8000) LastPixel = (NumPacks & 0xFF);
			if (i == 0xC000) {
				y += (65536 - NumPacks);
				if (y >= parentPtr->mHeader.mHeight) break;
			}
		}

		if (y >= parentPtr->mHeader.mHeight) break;
		dst = m8BitBuffer + y * mBytesPerLine;

		for (x = 0; NumPacks > 0; --NumPacks) {

			i = flicGetByte();
			x += i;
			if (x >= parentPtr->mHeader.mWidth) break;
			dst += i;

			i = flicGetByte();
			if (i > 127) {
				int pixel1 = flicGetByte();
				int pixel2 = flicGetByte();

				for ( ; i < 256; ++i) {
					if (x++ < parentPtr->mHeader.mWidth) {
						*dst++ = (unsigned char)pixel1;
					}
					if (x++ < parentPtr->mHeader.mWidth) {
						*dst++ = (unsigned char)pixel2;
					}
				}

			} else if (i > 0) {
				for (i <<= 1; i > 0; --i) {
					int pixel = flicGetByte();
					if (x++ < parentPtr->mHeader.mWidth) {
						*dst++ = (unsigned char)pixel;
					}
				}
			}
		}

		if (LastPixel != -1) {
			dst = m8BitBuffer + y * mBytesPerLine + parentPtr->mHeader.mWidth - 1;
			*dst = (unsigned char)LastPixel;
		}
	}
}

void VideoDecoderFLIC::flicLC() {
	int x, y;
	int NumLines, NumPacks;
	unsigned char *dst;

	y = flicGetWord();
	NumLines = flicGetWord();

	while (NumLines > 0) {

		if (y >= parentPtr->mHeader.mHeight) break;
		dst = m8BitBuffer + y * mBytesPerLine;

		NumPacks = flicGetByte();

		for (x = 0; NumPacks > 0; --NumPacks) {
			int i = flicGetByte();
			x += i;
			dst += i;
			i = flicGetByte();
			if (i > 127) {
				int pixel = flicGetByte();
				for (; i < 256; ++i) {
					if (x++ < parentPtr->mHeader.mWidth) {
						*dst++ = (unsigned char)pixel;
					}
				}
			} else {
				for ( ; i > 0; --i) {
					if (x++ < parentPtr->mHeader.mWidth) {
						*dst++ = (unsigned char)flicGetByte();
					}
				}
			}
		}

		++y;
		--NumLines;
	}

	return;
}

void VideoDecoderFLIC::flicBlack() {
	memset(m8BitBuffer, 0, mBytesPerLine * parentPtr->mHeader.mHeight);
}

void VideoDecoderFLIC::flicBrun() {
	int x, y;
	unsigned char *dst;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		dst = m8BitBuffer + y * mBytesPerLine;
		x = flicGetByte();	// ignore
		for (x = 0; x < parentPtr->mHeader.mWidth; ) {
			int i = flicGetByte();
			if (i > 127) {
				for ( ; i < 256; ++i) {
					if (x++ < parentPtr->mHeader.mWidth) {
						*dst++ = (unsigned char)flicGetByte();
					}
				}
			} else if (i > 0) {
				int pixel = flicGetByte();
				for ( ; i > 0; --i) {
					if (x++ < parentPtr->mHeader.mWidth) {
						*dst++ = (unsigned char)pixel;
					}
				}
			} else {
				// Don't endless loop!
				if (curbyte >= totbytes) {
					y = parentPtr->mHeader.mHeight;
					break;
				}
			}
		}
	}
}

void VideoDecoderFLIC::flicRaw() {
	int x, y;
	unsigned char *dst;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		dst = m8BitBuffer + y * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth; x > 0; --x) {
			*dst++ = (unsigned char)flicGetByte();
		}
	}
}

void VideoDecoderFLIC::flicConvertRGB15() {
	int x, y;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		unsigned short *dst = (unsigned short *)((char *)mPixmap.data + y * mPixmap.pitch);
		const unsigned char *src = m8BitBuffer + y * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth; x > 0; --x) {
			const int pixel = *src++;
			// 0rrrrrgg gggbbbbb
			*dst++ = (unsigned short)(
				((mPalette[pixel].rgbRed & 0xF8) << 7) |
				((mPalette[pixel].rgbGreen & 0xF8) << 2) |
				((mPalette[pixel].rgbBlue & 0xF8) >> 3)
			);
		}
	}
}

void VideoDecoderFLIC::flicConvertRGB16() {
	int x, y;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		unsigned short *dst = (unsigned short *)((char *)mPixmap.data + y * mPixmap.pitch);
		const unsigned char *src = m8BitBuffer + y * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth; x > 0; --x) {
			const int pixel = *src++;
			// rrrrrggg gggbbbbb
			*dst++ = (unsigned short)(
				((mPalette[pixel].rgbRed & 0xF8) << 8) |
				((mPalette[pixel].rgbGreen & 0xFC) << 3) |
				((mPalette[pixel].rgbBlue & 0xF8) >> 3)
			);
		}
	}
}

void VideoDecoderFLIC::flicConvertRGB24() {
	int x, y;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		unsigned char *dst = (unsigned char *)mPixmap.data + y * mPixmap.pitch;
		const unsigned char *src = m8BitBuffer + y * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth; x > 0; --x) {
			const int pixel = *src++;
			*dst++ = mPalette[pixel].rgbBlue;
			*dst++ = mPalette[pixel].rgbGreen;
			*dst++ = mPalette[pixel].rgbRed;
		}
	}
}

void VideoDecoderFLIC::flicConvertRGB32() {
	int x, y;
	const unsigned long *pal = (unsigned long *)mPalette;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		unsigned long *dst = (unsigned long *)((char *)mPixmap.data + y * mPixmap.pitch);
		const unsigned char *src = m8BitBuffer + y * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth; x > 0; --x) {
			*dst++ = pal[*src++];
		}
	}
}

void VideoDecoderFLIC::flicConvertY8() {
	int x, y;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		unsigned char *dst = (unsigned char *)((char *)mPixmap.data + y * mPixmap.pitch);
		const unsigned char *src = m8BitBuffer + y * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth; x > 0; --x) {
			*dst++ = yPalette[*src++].yuvY;
		}
	}
}

void VideoDecoderFLIC::flicConvertUYVY() {
	int x, y;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		unsigned long *dst = (unsigned long *)((char *)mPixmap.data + y * mPixmap.pitch);
		const unsigned char *src = m8BitBuffer + y * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth >> 1; x > 0; --x) {
			const YUVQUAD pixel1 = yPalette[*src++];
			const YUVQUAD pixel2 = yPalette[*src++];
			// Y V Y U
			*dst++ = (pixel2.yuvY << 24)
				| (((pixel1.yuvCr + pixel2.yuvCr + 1) >> 1) << 16)
				| (pixel1.yuvY << 8)
				| ((pixel1.yuvCb + pixel2.yuvCb + 1) >> 1);
		}
	}
}

void VideoDecoderFLIC::flicConvertYUY2() {
	int x, y;

	for (y = 0; y < parentPtr->mHeader.mHeight; ++y) {
		unsigned long *dst = (unsigned long *)((char *)mPixmap.data + y * mPixmap.pitch);
		const unsigned char *src = m8BitBuffer + y * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth >> 1; x > 0; --x) {
			const YUVQUAD pixel1 = yPalette[*src++];
			const YUVQUAD pixel2 = yPalette[*src++];
			// V Y U Y
			*dst++ = (((pixel1.yuvCr + pixel2.yuvCr + 1) >> 1) << 24)
				| (pixel2.yuvY << 16)
				| (((pixel1.yuvCb + pixel2.yuvCb + 1) >> 1) << 8)
				| pixel1.yuvY;
		}
	}
}

void VideoDecoderFLIC::flicConvertYV12() {
	int x, y;

	flicConvertY8();

	for (y = 0; y < (parentPtr->mHeader.mHeight >> 1); ++y) {
		unsigned char *dstU = (unsigned char *)((char *)mPixmap.data2 + y * mPixmap.pitch2);
		unsigned char *dstV = (unsigned char *)((char *)mPixmap.data3 + y * mPixmap.pitch3);
		const unsigned char *src = m8BitBuffer + y * 2 * mBytesPerLine;
		for (x = parentPtr->mHeader.mWidth >> 1; x > 0; --x) {
			const YUVQUAD pixel1 = yPalette[src[0]];
			const YUVQUAD pixel2 = yPalette[src[1]];
			const YUVQUAD pixel3 = yPalette[src[mBytesPerLine]];
			const YUVQUAD pixel4 = yPalette[src[mBytesPerLine + 1]];
			src += 2;
			*dstU++ = (unsigned char)((pixel1.yuvCb + pixel2.yuvCb + pixel3.yuvCb + pixel4.yuvCb + 2) >> 2);
			*dstV++ = (unsigned char)((pixel1.yuvCr + pixel2.yuvCr + pixel3.yuvCr + pixel4.yuvCr + 2) >> 2);
		}
	}
}

const void *VideoDecoderFLIC::DecodeFrame(
	const void *inputBuffer, uint32 data_len,
	bool is_preroll, sint64 streamFrame, sint64 targetFrame)
{
	unsigned long ChunkStart;
	long ChunkSize;
	int NumChunks;

	if (inputBuffer != NULL) {
		flicbytes = (unsigned char *)inputBuffer;
		totbytes = data_len;

		curbyte = 4;	// skip the frame size field
		if (flicGetWord() != 0xF1FA) goto Abort;

		NumChunks = flicGetWord();
		curbyte += 8;
		ChunkStart = curbyte;

		while (NumChunks > 0) {
			ChunkSize = flicGetDword();
			if (ChunkSize <= 0) goto Abort;
			if (ChunkSize & 1) ChunkSize++;

			switch (flicGetWord()) {
				case 4:		// COLOR_256
					flicPalette(false);
					break;
				case 7:		// DELTA_FLC
					flicDelta();
					break;
				case 11:	// COLOR_64
					flicPalette(true);
					break;
				case 12:	// DELTA_FLI
					flicLC();
					break;
				case 13:    // BLACK (Keyframe)
					flicBlack();
					break;
				case 15:    // BYTE_RUN (Keyframe)
					flicBrun();
					break;
				case 16:    // FLI_COPY (Keyframe)
					flicRaw();
					break;
			}

			curbyte = ChunkStart + ChunkSize;
			ChunkStart = curbyte;

			--NumChunks;
		}
	}

	if (!is_preroll) {
		mPrevFrame = (int)targetFrame;

		using namespace nsVDXPixmap;

		switch (mPixmap.format) {

//			case kPixFormat_Null:
			
			case kPixFormat_XRGB1555:
				flicConvertRGB15();
				break;

			case kPixFormat_RGB565:
				flicConvertRGB16();
				break;

			case kPixFormat_RGB888:
				flicConvertRGB24();
				break;

			case kPixFormat_XRGB8888:
				flicConvertRGB32();
				break;

			case kPixFormat_Y8:
				flicConvertY8();
				break;

			case kPixFormat_YUV422_UYVY:
				flicConvertUYVY();
				break;

			case kPixFormat_YUV422_YUYV:
				flicConvertYUY2();
				break;

//			case kPixFormat_YUV444_Planar:

//			case kPixFormat_YUV422_Planar:

			case kPixFormat_YUV420_Planar:
				flicConvertYV12();
				break;

//			case kPixFormat_YUV411_Planar:

//			case kPixFormat_YUV410_Planar:
		}
	}

Abort:
	return mFrameBuffer;
}

uint32 VideoDecoderFLIC::GetDecodePadding() {
	return 0;
}

void VideoDecoderFLIC::Reset() {
	mPrevFrame = -1;
}

bool VideoDecoderFLIC::IsFrameBufferValid() {
	return (mPrevFrame >= 0);
}

const VDXPixmap& VideoDecoderFLIC::GetFrameBuffer() {
	return mPixmap;
}

bool VideoDecoderFLIC::SetTargetFormat(int format, bool useDIBAlignment) {

	const int w = parentPtr->mHeader.mWidth;
	const int h = parentPtr->mHeader.mHeight;

	using namespace nsVDXPixmap;

	if (format == kPixFormat_Null) format = kPixFormat_RGB888;

	switch (format) {

//		case kPixFormat_Null:

		case kPixFormat_XRGB1555:

		case kPixFormat_RGB565:
			mPixmap.pitch		= (w * 2 + 3) & 0xFFFFFFFC;
			mPixmap.data		= mFrameBuffer + (h - 1) * mPixmap.pitch;
			mPixmap.pitch		= -mPixmap.pitch;
			mPixmap.palette		= NULL;
			mPixmap.w			= w;
			mPixmap.h			= h;
			mPixmap.format		= format;
			mPixmap.data2		= NULL;
			mPixmap.pitch2		= 0;
			mPixmap.data3		= NULL;
			mPixmap.pitch3		= 0;
			break;

		case kPixFormat_RGB888:
			mPixmap.pitch		= (w * 3 + 3) & 0xFFFFFFFC;
			mPixmap.data		= mFrameBuffer + (h - 1) * mPixmap.pitch;
			mPixmap.pitch		= -mPixmap.pitch;
			mPixmap.palette		= NULL;
			mPixmap.w			= w;
			mPixmap.h			= h;
			mPixmap.format		= format;
			mPixmap.data2		= NULL;
			mPixmap.pitch2		= 0;
			mPixmap.data3		= NULL;
			mPixmap.pitch3		= 0;
			break;

		case kPixFormat_XRGB8888:
			mPixmap.data		= mFrameBuffer + (h - 1) * w * 4;
			mPixmap.palette		= NULL;
			mPixmap.w			= w;
			mPixmap.h			= h;
			mPixmap.pitch		= w * -4;
			mPixmap.format		= format;
			mPixmap.data2		= NULL;
			mPixmap.pitch2		= 0;
			mPixmap.data3		= NULL;
			mPixmap.pitch3		= 0;
			break;

		case kPixFormat_Y8:
			mPixmap.data		= mFrameBuffer;
			mPixmap.pitch		= w;
			mPixmap.palette		= NULL;
			mPixmap.format		= format;
			mPixmap.w			= w;
			mPixmap.h			= h;
			mPixmap.data2		= NULL;
			mPixmap.pitch2		= 0;
			mPixmap.data3		= NULL;
			mPixmap.pitch3		= 0;
			break;

		case kPixFormat_YUV422_UYVY:

		case kPixFormat_YUV422_YUYV:
			if (w & 1) return false;
			mPixmap.pitch		= w * 2;
			mPixmap.data		= mFrameBuffer;
			mPixmap.palette		= NULL;
			mPixmap.w			= w;
			mPixmap.h			= h;
			mPixmap.format		= format;
			mPixmap.data2		= NULL;
			mPixmap.pitch2		= 0;
			mPixmap.data3		= NULL;
			mPixmap.pitch3		= 0;
			break;

//		case kPixFormat_YUV444_Planar:

//		case kPixFormat_YUV422_Planar:

		case kPixFormat_YUV420_Planar:
			if ((w | h) & 1) return false;
			mPixmap.data		= mFrameBuffer;
			mPixmap.palette		= NULL;
			mPixmap.format		= format;
			mPixmap.w			= w;
			mPixmap.h			= h;
			mPixmap.pitch		= w;
			mPixmap.data2		= mFrameBuffer + w * h;
			mPixmap.pitch2		= w >> 1;
			mPixmap.data3		= mFrameBuffer + w * h + (w >> 1) * (h >> 1);
			mPixmap.pitch3		= w >> 1;
			break;

//		case kPixFormat_YUV411_Planar:

//		case kPixFormat_YUV410_Planar:

		default:
			return false;
	}


	Reset();

	return true;
}

bool VideoDecoderFLIC::SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih) {
	return false;
}

const void *VideoDecoderFLIC::GetFrameBufferBase() {
	return mFrameBuffer;
}

bool VideoDecoderFLIC::IsDecodable(sint64 sample_num64) {
	if (parentPtr->mIndex[(uint32)sample_num64].mSize < 0) return true;
	if ((int)(sample_num64 - 1) == mPrevFrame) return true;
	if ((int)sample_num64 == mPrevFrame) return true;
	return false;
}


///////////////////////////////////////////////////////////////////////
//
//						VideoDecoderModelFLIC
//
///////////////////////////////////////////////////////////////////////

#ifndef USE_DEFAULT_IP_MODEL

class VideoDecoderModelFLIC : public vdxunknown<IVDXVideoDecoderModel> {
public:
	VideoDecoderModelFLIC(InputFileFLIC *const pp);
	~VideoDecoderModelFLIC();

	void	VDXAPIENTRY Reset();
	void	VDXAPIENTRY SetDesiredFrame(sint64 frame_num);
	sint64	VDXAPIENTRY GetNextRequiredSample(bool& is_preroll);
	int		VDXAPIENTRY GetRequiredCount();

private:
	sint64	mLastFrame;
	sint64	mTargetFrame;
	sint64	mNextFrame;

	InputFileFLIC *const parentPtr;
};

VideoDecoderModelFLIC::VideoDecoderModelFLIC(InputFileFLIC *const pp)
	: parentPtr(pp)
{
	parentPtr->AddRef();
}

VideoDecoderModelFLIC::~VideoDecoderModelFLIC() {
	parentPtr->Release();
}

void VideoDecoderModelFLIC::Reset() {
	mLastFrame = -1;
}

void VideoDecoderModelFLIC::SetDesiredFrame(sint64 frame_num) {
	mTargetFrame = frame_num;

	// Do we already have the frame?
	if (mLastFrame == frame_num) {
		mNextFrame = -1;	// Flag this for GetNextRequiredSample
		return;
	}

	mNextFrame = frame_num;

	// If it's a Delta frame, locate a previous Key
	if (parentPtr->mIndex[(uint32)frame_num].mSize < 0) {

		while (mNextFrame > 0) {
			--mNextFrame;
			if (mNextFrame == mLastFrame) {
				// Start from the frame following that one
				++mNextFrame;
				break;
			}
			if (parentPtr->mIndex[(uint32)mNextFrame].mSize < 0) break;
		}
	}
}

sint64 VideoDecoderModelFLIC::GetNextRequiredSample(bool& is_preroll) {
	sint64 frame;

	if (mNextFrame == mTargetFrame || mNextFrame < 0) {
		is_preroll = false;
		mLastFrame = mTargetFrame;
		frame = mNextFrame;
		mNextFrame = -1;
	} else {
		is_preroll = true;
		frame = mNextFrame++;
	}

	return frame;
}

int VideoDecoderModelFLIC::GetRequiredCount() {
	if (mNextFrame < 0) return 0;

	return (int)((mTargetFrame - mNextFrame) + 1);
}

#endif	// USE_DEFAULT_IP_MODEL


///////////////////////////////////////////////////////////////////////
//
//						VideoSourceFLIC
//
///////////////////////////////////////////////////////////////////////

class VideoSourceFLIC : public vdxunknown<IVDXStreamSource>, public IVDXVideoSource {
public:
	VideoSourceFLIC(InputFileFLIC *const pp);
	~VideoSourceFLIC();

	int VDXAPIENTRY AddRef();
	int VDXAPIENTRY Release();
	void *VDXAPIENTRY AsInterface(uint32 iid);

	void		VDXAPIENTRY GetStreamSourceInfo(VDXStreamSourceInfo&);
	bool		VDXAPIENTRY Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);

	const void *VDXAPIENTRY GetDirectFormat();
	int			VDXAPIENTRY GetDirectFormatLen();

	ErrorMode VDXAPIENTRY GetDecodeErrorMode();
	void VDXAPIENTRY SetDecodeErrorMode(ErrorMode mode);
	bool VDXAPIENTRY IsDecodeErrorModeSupported(ErrorMode mode);

	bool VDXAPIENTRY IsVBR();
	sint64 VDXAPIENTRY TimeToPositionVBR(sint64 us);
	sint64 VDXAPIENTRY PositionToTimeVBR(sint64 samples);

	void VDXAPIENTRY GetVideoSourceInfo(VDXVideoSourceInfo& info);

	bool VDXAPIENTRY CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel);
	bool VDXAPIENTRY CreateVideoDecoder(IVDXVideoDecoder **ppDecoder);

	void		VDXAPIENTRY GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo);

	bool		VDXAPIENTRY IsKey(sint64 lSample);

	sint64		VDXAPIENTRY GetFrameNumberForSample(sint64 sample_num);
	sint64		VDXAPIENTRY GetSampleNumberForFrame(sint64 display_num);
	sint64		VDXAPIENTRY GetRealFrame(sint64 display_num);

	sint64		VDXAPIENTRY GetSampleBytePosition(sint64 sample_num);

private:
	InputFileFLIC *const parentPtr;

	FLICINFOHEADER fih;
};

VideoSourceFLIC::VideoSourceFLIC(InputFileFLIC *const pp)
	: parentPtr(pp)
{
	parentPtr->AddRef();

	fih.bmih.biSize = sizeof(BITMAPINFOHEADER);
	fih.bmih.biWidth = parentPtr->mHeader.mWidth;
	fih.bmih.biHeight = parentPtr->mHeader.mHeight;
	fih.bmih.biPlanes = 1;
	fih.bmih.biBitCount = 8;
	fih.bmih.biCompression = BI_RGB;	// ' ELR';
	fih.bmih.biSizeImage = 0;
	fih.bmih.biXPelsPerMeter = 0;
	fih.bmih.biYPelsPerMeter = 0;
	fih.bmih.biClrUsed = 0;
	fih.bmih.biClrImportant = 0;

	for (int i = 0; i < 256; ++i) {
		fih.bmic[i].rgbRed = i;
		fih.bmic[i].rgbGreen = i;
		fih.bmic[i].rgbBlue = i;
		fih.bmic[i].rgbReserved = 255;
	}
}

VideoSourceFLIC::~VideoSourceFLIC() {
	parentPtr->Release();
}

int VideoSourceFLIC::AddRef() {
	return vdxunknown<IVDXStreamSource>::AddRef();
}

int VideoSourceFLIC::Release() {
	return vdxunknown<IVDXStreamSource>::Release();
}

void *VDXAPIENTRY VideoSourceFLIC::AsInterface(uint32 iid) {
	if (iid == IVDXVideoSource::kIID)
		return static_cast<IVDXVideoSource *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}

void VDXAPIENTRY VideoSourceFLIC::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo) {
	if (parentPtr->mHeader.mVersion == 0xAF11) {
		// FLI: frame rate is in jiffies (1/70 second)
		srcInfo.mSampleRate.mNumerator = 70;
	} else {
		// FLC: frame rate is in milliseconds
		srcInfo.mSampleRate.mNumerator = 1000;
	}
	srcInfo.mSampleRate.mDenominator = parentPtr->mHeader.mFrameRate;
	srcInfo.mSampleCount = parentPtr->mHeader.mFrames;
}

bool VideoSourceFLIC::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	uint32 uStart = (uint32)lStart64;

	if (uStart >= (uint32)parentPtr->mHeader.mFrames) {
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead) *lBytesRead = 0;
		return true;
	}

	const FlicIndexEntry *fie = &(parentPtr->mIndex[uStart]);
	long sz = fie->mSize & 0x7FFFFFFF;

	if (lpBuffer == NULL || cbBuffer < (uint32)sz) {
		if (lSamplesRead) *lSamplesRead = 1;
		if (lBytesRead) *lBytesRead = sz;
		return (lpBuffer == NULL);
	}

	fseek(parentPtr->mFile, fie->mFrameOffset, SEEK_SET);
	fread(lpBuffer, sz, 1, parentPtr->mFile);

	if (lSamplesRead) *lSamplesRead = 1;
	if (lBytesRead) *lBytesRead = sz;
	return true;
}

const void *VideoSourceFLIC::GetDirectFormat() {
	return NULL;
}

int VideoSourceFLIC::GetDirectFormatLen() {
	return 0;
}

IVDXStreamSource::ErrorMode VideoSourceFLIC::GetDecodeErrorMode() {
	return IVDXStreamSource::kErrorModeReportAll;
}

void VideoSourceFLIC::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode) {
}

bool VideoSourceFLIC::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode) {
	return mode == IVDXStreamSource::kErrorModeReportAll;
}

bool VideoSourceFLIC::IsVBR() {
	return false;
}

sint64 VideoSourceFLIC::TimeToPositionVBR(sint64 us) {
	double mNumerator;

	if (parentPtr->mHeader.mVersion == 0xAF11) {
		mNumerator = 70;
	} else {
		mNumerator = 1000;
	}

	return (sint64)(0.5 + us / 1000000.0 *
		mNumerator / (double)parentPtr->mHeader.mFrameRate);
}

sint64 VideoSourceFLIC::PositionToTimeVBR(sint64 samples) {
	double mNumerator;

	if (parentPtr->mHeader.mVersion == 0xAF11) {
		mNumerator = 70;
	} else {
		mNumerator = 1000;
	}

	return (sint64)(0.5 + samples * 1000000.0 *
		(double)parentPtr->mHeader.mFrameRate / mNumerator);
}

void VideoSourceFLIC::GetVideoSourceInfo(VDXVideoSourceInfo& info) {
	info.mFlags = 0;
	info.mWidth = parentPtr->mHeader.mWidth;
	info.mHeight = parentPtr->mHeader.mHeight;

#ifdef USE_DEFAULT_IP_MODEL
	info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
#else
	info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelCustom;
#endif
}

bool VideoSourceFLIC::CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel) {
	*ppModel = NULL;

#ifdef USE_DEFAULT_IP_MODEL
	return false;
#else
	IVDXVideoDecoderModel *p = new VideoDecoderModelFLIC(parentPtr);
	if (!p) return false;

	p->AddRef();
	*ppModel = p;
	return true;
#endif
}

bool VideoSourceFLIC::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder) {
	*ppDecoder = NULL;

	IVDXVideoDecoder *p = new VideoDecoderFLIC(parentPtr);
	if (!p)	return false;

	p->AddRef();
	*ppDecoder = p;
	return true;
}

void VideoSourceFLIC::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo) {
	frameInfo.mBytePosition = parentPtr->mIndex[(uint32)sample_num].mFrameOffset;

	if (parentPtr->mIndex[(uint32)sample_num].mSize < 0) {
		frameInfo.mFrameType = kVDXVFT_Independent;
		frameInfo.mTypeChar = 'K';
	} else {
		frameInfo.mFrameType = kVDXVFT_Predicted;
		frameInfo.mTypeChar = ' ';
	}
}

bool VideoSourceFLIC::IsKey(sint64 sample) {
	return (parentPtr->mIndex[(uint32)sample].mSize < 0);
}

sint64 VideoSourceFLIC::GetFrameNumberForSample(sint64 sample_num) {
	return sample_num;
}

sint64 VideoSourceFLIC::GetSampleNumberForFrame(sint64 display_num) {
	return display_num;
}

sint64 VideoSourceFLIC::GetRealFrame(sint64 display_num) {
	return display_num;
}

sint64 VideoSourceFLIC::GetSampleBytePosition(sint64 sample_num) {
	return parentPtr->mIndex[(uint32)sample_num].mFrameOffset;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileFLIC
//
///////////////////////////////////////////////////////////////////////

InputFileFLIC::InputFileFLIC(const VDXInputDriverContext& context)
	: mContext(context)
{
	mIndex = NULL;
	mFile = NULL;
}

InputFileFLIC::~InputFileFLIC()
{
	free(mIndex);
	if (mFile != NULL) fclose(mFile);
}

void InputFileFLIC::Init(const wchar_t *szFile, IVDXInputOptions *opts) {
	int i, NumChunks;
	unsigned long ChunkStart, ChunkSize;
	unsigned long frame_origin;

	if (GetVersion() & 0x80000000) {
		// Windows 95, 98
		i = WideCharToMultiByte(CP_ACP, 0, szFile, -1, NULL, 0, NULL, NULL);

		if (i > 0) {
			char *tmp = (char *)malloc(i + 1);

			if (tmp != NULL) {
				if (WideCharToMultiByte(CP_ACP, 0, szFile, -1, tmp, i, NULL, NULL) > 0) {
					mFile = fopen(tmp, "rb");
				}
				free(tmp);
			}
		}

	} else {
		// Windows NT, XP
		mFile = _wfopen(szFile, L"rb");
	}

	if (mFile == NULL) {
		mContext.mpCallbacks->SetError("Impossible d'ouvrir le fichier: %ls", szFile);
		return;
	}

	fread(&mHeader, sizeof(mHeader), 1, mFile);

	if (mHeader.mFrames <= 0) {
		mContext.mpCallbacks->SetError("Aucune image vido n'a t trouve");
		return;
	}

	if (mHeader.mVersion == 0xAF11) {
		if (mHeader.mWidth != 320 || mHeader.mHeight != 200) {
			mHeader.mWidth = 0;
		}
	} else {
		if (mHeader.mWidth <= 0 || mHeader.mHeight <= 0) {
			mHeader.mWidth = 0;
		}
	}

	if (mHeader.mWidth == 0) {
		mContext.mpCallbacks->SetError("Largeur / hauteur invalide");
		return;
	}

	// Do not allow a frame rate of zero
	if (mHeader.mFrameRate == 0) {
		if (mHeader.mVersion == 0xAF11) {
			mHeader.mFrameRate = 4;
		} else {
			mHeader.mFrameRate = 57;
		}
	}

	mIndex = (FlicIndexEntry *)malloc(
		mHeader.mFrames * sizeof(FlicIndexEntry));

	if (mIndex == NULL) {
		mContext.mpCallbacks->SetErrorOutOfMemory();
		return;
	}

	frame_origin = 128;

	for (i = 0; i < mHeader.mFrames; ++i) {

		fseek(mFile, frame_origin, SEEK_SET);

		mIndex[i].mFrameOffset = frame_origin;
		fread(&(mIndex[i].mSize), 4, 1, mFile);

		NumChunks = 0;
		fseek(mFile, frame_origin + 6, SEEK_SET);
		fread(&NumChunks, 2, 1, mFile);

		ChunkStart = frame_origin + 8 + 8;

		while (NumChunks > 0) {
			fseek(mFile, ChunkStart, SEEK_SET);
			fread(&ChunkSize, 4, 1, mFile);
			if (ChunkSize) {
				int t = 0;
				if (ChunkSize & 1) ++ChunkSize;

				fread(&t, 2, 1, mFile);
				if (t == 13 || t == 15 || t == 16) {
					// Frame is a Keyframe
					mIndex[i].mSize |= 0x80000000;
					break;
				}
			}

			ChunkStart += ChunkSize;
			--NumChunks;
		}

		frame_origin += (mIndex[i].mSize & 0x7FFFFFFF);
	}
}

bool InputFileFLIC::Append(const wchar_t *szFile) {
	return false;
}

bool InputFileFLIC::PromptForOptions(VDXHWND, IVDXInputOptions **ppOptions) {
	*ppOptions = NULL;
	return false;
}

bool InputFileFLIC::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions) {
	*ppOptions = NULL;
	return false;
}

INT_PTR CALLBACK InputFileFLIC::InfoDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	InputFileFLIC *thisPtr;
	double fps;
	char buf[32];

	switch (message) {

    case WM_INITDIALOG:
		thisPtr = (InputFileFLIC *)lParam;

		sprintf(buf, "0x%04X", thisPtr->mHeader.mVersion);
		SetDlgItemText(hDlg, IDC_VERSION, buf);

		sprintf(buf, "%i x %i", thisPtr->mHeader.mWidth, thisPtr->mHeader.mHeight);
		SetDlgItemText(hDlg, IDC_FRAME_SIZE, buf);

		if (thisPtr->mHeader.mVersion == 0xAF11) {
			// FLI: frame rate is in jiffies (1/70 second)
			sprintf(buf, "%i jiffies", thisPtr->mHeader.mFrameRate);
			fps = 70;
		} else {
			// FLC: frame rate is in milliseconds
			sprintf(buf, "%i millisecondes", thisPtr->mHeader.mFrameRate);
			fps = 1000;
		}
		SetDlgItemText(hDlg, IDC_FRAME_DELAY, buf);
		fps = fps / thisPtr->mHeader.mFrameRate;

		sprintf(buf, "%.3f fps", fps);
		SetDlgItemText(hDlg, IDC_FRAME_RATE, buf);

		return TRUE;

	case WM_COMMAND:
		// Note: "IDCANCEL" is the [X] button
		// Return TRUE if we process the message

		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
			EndDialog(hDlg, TRUE);
			return TRUE;
		}
		break;

    }

    return FALSE;
}


void InputFileFLIC::DisplayInfo(VDXHWND hwndParent) {
    DialogBoxParam(g_hModule, MAKEINTRESOURCE(IDD_INFORMATION),
		(HWND)hwndParent, InfoDlgProc, (LPARAM)this);
}

bool InputFileFLIC::GetVideoSource(int index, IVDXVideoSource **ppVS) {
	*ppVS = NULL;

	if (index) return false;

	IVDXVideoSource *pVS = new VideoSourceFLIC(this);
	if (!pVS) return false;

	*ppVS = pVS;
	pVS->AddRef();
	return true;
}

bool InputFileFLIC::GetAudioSource(int index, IVDXAudioSource **ppAS) {
	*ppAS = NULL;
	return false;
}

///////////////////////////////////////////////////////////////////////////////

class InputFileDriverFLIC : public vdxunknown<IVDXInputFileDriver> {
public:
	InputFileDriverFLIC(const VDXInputDriverContext& context);

	int		VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool	VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	int		mRefCount;

	const VDXInputDriverContext& mContext;
};

InputFileDriverFLIC::InputFileDriverFLIC(const VDXInputDriverContext& context)
	: mRefCount(0)
	, mContext(context)
{
}

int VDXAPIENTRY InputFileDriverFLIC::DetectBySignature(
	const void *pHeader, sint32 nHeaderSize,
	const void *pFooter, sint32 nFooterSize, sint64 nFileSize)
{
	return -1;
}

bool VDXAPIENTRY InputFileDriverFLIC::CreateInputFile(uint32 flags, IVDXInputFile **ppFile) {
	*ppFile = NULL;

	IVDXInputFile *p = new InputFileFLIC(mContext);
	if (!p) return false;

	*ppFile = p;
	p->AddRef();
	return true;
}

bool VDXAPIENTRY flic_create(const VDXInputDriverContext *pContext, IVDXInputFileDriver **ppDriver) {
	*ppDriver = NULL;

	IVDXInputFileDriver *p = new InputFileDriverFLIC(*pContext);
	if (!p) return false;

	*ppDriver = p;
	p->AddRef();
	return true;
}

const uint8 FLIC_sig[] = {
	0, 0, 0, 0, 0, 0, 0, 0,		// filesize (ignored)
	0x10, 0xFC, 0xAF, 0xFF,		// 0xAF11 or 0xAF12 (version)
	0, 0, 0, 0x80,				// numframes (can't be negative)
	0, 0, 0, 0x80,				// width (can't be negative)
	0, 0, 0, 0x80,				// height (can't be negative)
	0x08, 0xFF, 0x00, 0xFF		// bits per pixel (must be 8)
};

const VDXInputDriverDefinition flic_input = {
	sizeof(VDXInputDriverDefinition),		// size of this structure
	VDXInputDriverDefinition::kFlagSupportsVideo,
	0,										// priority
	sizeof(FLIC_sig),						// size of file signature
	FLIC_sig,								// pointer to signature
	L"*.fli|*.flc",							// extension pattern
	L"Fichiers FLIC (*.fli,*.flc)|*.fli;*.flc",	// description pattern
	L"FLIC",								// driver target name
	flic_create								// create proc
};

const VDXPluginInfo flic_info = {
	sizeof(VDXPluginInfo),				// size of this structure
	L"FLIC",							// name of this plugin
	L"fccHandler",						// author's name
	L"Charge des animations FLIC.",			// plugin description
	0x01000000,							// version = 1.0.0
	kVDXPluginType_Input,				// type of plugin = input
	0,									// flags = 0
	10,									// API version required
	kVDXPlugin_APIVersion,				// API version used
	2,									// API version required
	kVDXPlugin_InputDriverAPIVersion,	// API version used
	&flic_input							// driver definition
};

///////////////////////////////////////////////////////////////////////////

const VDXPluginInfo *const kPlugins[] = {
	&flic_info,
	NULL
};

extern "C" __declspec(dllexport) const VDXPluginInfo *const * VDXAPIENTRY VDGetPluginInfo() {
	return kPlugins;
}
